"""
2021.12.01 Yixuan Mei
This file contains class Server, which manages SharedMemoryObjectStore in a remote process.
"""

import multiprocessing as mp
import time
from multiprocessing.managers import BaseManager

import SMOS_exceptions
import SMOS_utils as utils
from SMOS_constants import SMOS_SUCCESS
from SMOS_shared_memory_object_store import SharedMemoryObjectStore


class Server:

    def __init__(self, connection: utils.ConnectionDescriptor = None):
        """
        Each Server instance runs an instance of SharedMemoryObjectStore remotely.

        :param connection: (optional) a ConnectionDescriptor that specifies address of SMOS server
        """
        if connection is not None:
            self.connection = connection
        else:
            port = utils.get_local_free_port(1, 5000, 5050)[0]
            self.connection = utils.ConnectionDescriptor(ip="localhost", port=port,
                                                         authkey=b"antony")
        self.server_process = None
        self.object_store = None

    def address(self):
        """
        Get the address that server listens

        :return: address that server listens
        """
        return self.connection

    def start(self):
        """
        Start server in another process.

        :return: always SMOS_SUCCESS
        """
        self.server_process = mp.Process(target=start_server, args=(self.connection,))
        self.server_process.start()
        self.object_store = get_object_store(self.connection)
        return SMOS_SUCCESS

    def stop(self):
        """
        Stops server for safe exit.

        :return: always SMOS_SUCCESS
        """
        self.object_store.stop()
        self.server_process.terminate()
        return SMOS_SUCCESS


class SMOSManager(BaseManager):
    """
    Customized multiprocessing manager for SMOS.
    """
    pass


def start_server(connection: utils.ConnectionDescriptor):
    """
    Registers SMOS to python multiprocessing manager and start to serve.
    Note that this function must be called as a separate process.

    :exception SMOS_exceptions.SMOSServerDropOut: if server drops out by accident

    :param connection: a ConnectionDescriptor that specifies the address that SMOS server listens
    """
    # register SMOS
    object_store = SharedMemoryObjectStore()
    SMOSManager.register("get_object_store", callable=lambda: object_store)
    utils.log2terminal(info_type="Info", msg="Server registered.")

    # run server
    manager = SMOSManager(address=(connection.ip, connection.port), authkey=connection.authkey)
    server = manager.get_server()
    utils.log2terminal(info_type="Info", msg="Server started.")
    server.serve_forever()

    # we can never reach here
    raise SMOS_exceptions.SMOSServerDropOut("Server drops")


def get_object_store(connection: utils.ConnectionDescriptor) -> SharedMemoryObjectStore:
    """
    Get a reference of SMOS that lives in server process.

    :param connection: a ConnectionDescriptor that specifies the address that SMOS server listens
    :return: a reference of SMOS
    """
    # register SMOS
    SMOSManager.register("get_object_store")
    manager = SMOSManager(address=(connection.ip, connection.port), authkey=connection.authkey)

    # connect to server
    connected = False
    while not connected:
        try:
            manager.connect()
            connected = True
        except ConnectionRefusedError:
            utils.log2terminal(info_type="Info", msg="Waiting for server to start...")
            time.sleep(1)

    # return reference
    return manager.get_object_store()



"""
2021.11.24 Yixuan Mei
This file contains exceptions in shared memory object store.
"""


class SMOSPortBusy(Exception):
    """
    This exception is raised when there are not enough free ports available.
    """
    pass


class SMOSInputTypeError(Exception):
    """
    This exception is raised when input type is not the same as expected.
    """
    pass


class SMOSEntryUnallocated(Exception):
    """
    This exception is raised when user tries to append an unallocated block
    into data tracks.
    """
    pass


class SMOSReadRefDoubleRelease(Exception):
    """
    This exception is raised when read reference to an entry is released multiple
    times. Note that this exception may not be raised exactly when double release
    happens, but a double release is sure to cause this exception in itself or
    later releases.
    """
    pass


class SMOSBlockDoubleRelease(Exception):
    """
    This exception is raised when the block to be freed is already in free block
    list when deleting an entry form a data track. This is probably caused by
    erroneously mapping two entries to the same block or calling free_block_mapping
    twice.
    """
    pass


class SMOSTrackMismatch(Exception):
    """
    This exception is raised when track_name of current track is different from input
    entry_config's track_name.
    """
    pass


class SMOSDimensionMismatch(Exception):
    """
    This exception is raised when multiple objects that are supposed to have the same
    shape turn out to be different.
    """
    pass


class SMOSTrackUnaligned(Exception):
    """
    This exception is raised when multiple tracks in the same SharedMemoryObject are
    unaligned, i.e. they have different remaining block count.
    """
    pass


class SMOSMappingError(Exception):
    """
    This exception is raised when SMOS finds that an entry is mapped to a out-of-range
    block in some DataTrack.
    """
    pass


class SMOSObjectExistError(Exception):
    """
    This exception is raised when trying to create a new SharedMemoryObject using name
    that already exists in SMOS.
    """
    pass


class SMOSObjectNotFoundError(Exception):
    """
    This exception is raised when trying a access a SharedMemoryObject that does not
    exist in SMOS.
    """
    pass


class SMOSServerDropOut(Exception):
    """
    This exception is raised when SMOS server process accidentally terminates.
    """
    pass



"""
2021.11.25 Yixuan Mei
This file contains definition of class DataTrack for Shared Memory Object Store
"""

import queue
import random
from multiprocessing import shared_memory

import SMOS_exceptions
import SMOS_utils as utils
from SMOS_constants import SMOS_FAIL, SMOS_SUCCESS, SMOS_PERMISSION_DENIED, SMOS_MAX


class DataTrack:

    def __init__(self, track_name, shm_name, block_size, max_capacity):
        """
        Each data track manages a shared memory space. Can be used as a buffer or queue.
        Note that data tracks are components of SharedMemoryObject. Therefore, we do not
        handle concurrency issues at this level. All locks are imposed on SharedMemoryObject.

        :param track_name: name of this data track
        :param shm_name: name of the underlying shared memory space
        :param block_size: size of each block
        :param max_capacity: maximum number of objects that can be stored in this track
        """
        # parameters
        self.track_name = track_name
        self.shm_name = shm_name
        self.block_size = block_size
        self.max_capacity = max_capacity

        # underlying shared memory space
        self.shm = shared_memory.SharedMemory(create=True, size=block_size * max_capacity, name=shm_name)

        # array management
        self.entry_config_list = {}
        self.next_key = 0
        self.free_block_list = queue.Queue(maxsize=max_capacity)
        for i in range(max_capacity):
            self.free_block_list.put(i)

    # write
    def allocate_block(self, entry_config: utils.EntryConfig):
        """
        Allocate a free block for a new entry and write into entry config.

        :param entry_config: configuration of new entry
        :return: SMOS_SUCCESS if successful,
                 SMOS_FAIL if no free block available
        """
        try:
            block_idx = self.free_block_list.get(block=False)
            entry_config.mapped_block_idx = block_idx
            entry_config.track_name = self.track_name
            return SMOS_SUCCESS
        except queue.Empty:
            return SMOS_FAIL

    def append_entry_config(self, entry_config: utils.EntryConfig):
        """
        Append configuration of new entry to this data track's configuration list

        :exception object_store_exceptions.SMOSEntryUnallocated: if entry_config is mapped to -1 block
        :exception object_store_exceptions.SMOSTrackMismatch: if track_name of current track is different
                   from track name of input entry_config

        :param entry_config: configuration of new entry
        :return: always [SMOS_SUCCESS, index of appended entry config]
        """
        # check if entry_config has been correctly allocated
        if entry_config.mapped_block_idx == -1:
            raise SMOS_exceptions.SMOSEntryUnallocated(f"Entry unallocated.")
        if not entry_config.track_name == self.track_name:
            raise SMOS_exceptions.SMOSTrackMismatch(f"Current track is {self.track_name}, while "
                                                    f"input entry_config is associated with track"
                                                    f" {entry_config.track_name}.")

        # append entry config
        self.entry_config_list[self.next_key] = entry_config
        self.next_key += 1
        return SMOS_SUCCESS, self.next_key - 1

    # read
    def read_entry_config(self, idx):
        """
        Read entry config at given index and add read reference to that entry.

        :param idx: index of entry to be read
        :return: [SMOS_SUCCESS, entry_config] if successful,
                 [SMOS_FAIL, None] if target entry does not exist
        """
        try:
            self.entry_config_list[idx].pending_readers += 1
            entry_config = self.entry_config_list[idx]
            return SMOS_SUCCESS, entry_config
        except KeyError:
            return SMOS_FAIL, None

    def release_read_reference(self, idx):
        """
        Release read reference on given entry.

        :exception object_store_exceptions.SMOSReadRefDoubleRelease: if a read reference is
                   released multiple times.

        :param idx: index of entry to be released
        :return: SMOS_SUCCESS if successful,
                 SMOS_FAIL if target entry does not exist
        """
        try:
            self.entry_config_list[idx].pending_readers -= 1
            if self.entry_config_list[idx].pending_readers < 0:
                raise SMOS_exceptions.SMOSReadRefDoubleRelease(f"Double release on track {self.track_name}"
                                                               f"index {idx}")
            return SMOS_SUCCESS
        except KeyError:
            return SMOS_FAIL

    # delete
    def delete_entry_config(self, idx, force_delete=False):
        """
        Delete an entry from current data track. Note that this is lazy delete, the actual
        data in shared memory is not erased.

        :exception object_store_exceptions.SMOSBlockDoubleRelease: if block to which the deleted
        entry is mapped has already been freed.

        :param idx: index of entry to be deleted
        :param force_delete: whether to delete the entry when there are still pending readers
        :return: SMOS_SUCCESS if successful,
                 SMOS_FAIL if target entry does not exist,
                 SMOS_PERMISSION_DENIED if permission denied
        """
        try:
            # check delete permission
            delete_permission = (self.entry_config_list[idx].pending_readers == 0)
            if not delete_permission and not force_delete:
                return SMOS_PERMISSION_DENIED

            # delete entry config and free
            entry_config = self.entry_config_list[idx]
            block_idx = entry_config.mapped_block_idx
            del self.entry_config_list[idx]
            if block_idx in self.free_block_list.queue:
                raise SMOS_exceptions.SMOSBlockDoubleRelease(f"Block {block_idx} has already been"
                                                             f"freed in data track {self.track_name}.")
            else:
                self.free_block_list.put(block_idx)
                return SMOS_SUCCESS

        except KeyError:
            return SMOS_FAIL

    # pop and free
    def pop_entry_config(self, force_pop=False):
        """
        Pop an entry from current data track. Note that block to which the entry is mapped
        will not be freed in this function (since data in the entry will be used after pop).
        Call free_block_mapping to free the block when data in the entry is no longer useful.

        :param force_pop: whether to pop the entry when there are still pending readers
        :return: [SMOS_SUCCESS, entry_config] if successful,
                 [SMOS_FAIL, None] if data track empty,
                 [SMOS_PERMISSION_DENIED, None] if permission denied
        """
        try:
            # get smallest item
            entry_idx = min(list(self.entry_config_list.keys()))

            # check permission
            pop_permission = (self.entry_config_list[entry_idx].pending_readers == 0)
            if not pop_permission and not force_pop:
                return SMOS_PERMISSION_DENIED, None

            # pop entry config
            entry_config = self.entry_config_list[entry_idx]
            del self.entry_config_list[entry_idx]
            return SMOS_SUCCESS, entry_config

        except ValueError:
            return SMOS_FAIL, None

    def free_block_mapping(self, entry_config: utils.EntryConfig):
        """
        Free a block associated with a previously popped entry.

        :exception object_store_exceptions.SMOSBlockDoubleRelease: if the block associated with
                   input entry_config has already been freed
        :exception object_store_exceptions.SMOSTrackMismatch: if track_name of current track is different
                   from track name of input entry_config

        :param entry_config: a previously popped entry
        :return: always SMOS_SUCCESS
        """
        # check if input entry_config is associated with current track
        if not entry_config.track_name == self.track_name:
            raise SMOS_exceptions.SMOSTrackMismatch(f"Current track is {self.track_name}, while "
                                                    f"input entry_config is associated with track"
                                                    f" {entry_config.track_name}.")

        # free block mapping
        block_idx = entry_config.mapped_block_idx
        if block_idx in self.free_block_list.queue:
            raise SMOS_exceptions.SMOSBlockDoubleRelease(f"Block {block_idx} has already been"
                                                         f"freed in data track {self.track_name}.")
        else:
            self.free_block_list.put(block_idx)
            return SMOS_SUCCESS

    # stop track for clean up
    def stop(self):
        """
        Unlinks current track's shared memory. This function is supposed to
        be called when (and only when) cleaning up for exit.

        :return: always SMOS_SUCCESS
        """
        self.shm.close()
        self.shm.unlink()
        return SMOS_SUCCESS

    # utility functions for DataTrack
    def get_entry_count(self):
        """
        Get the number of entries in current track.

        :return: entry count
        """
        return len(self.entry_config_list)

    def get_entry_offset(self, entry_config: utils.EntryConfig):
        """
        Get offset of given entry in shared memory space.

        :exception object_store_exceptions.SMOSTrackMismatch: if track_name of current track is different
                   from track name of input entry_config
        :exception SMOS_exceptions.SMOSMappingError: if mapped_block_idx in query is not smaller than
                   max_capacity

        :param entry_config: entry to be queried
        :return: always [SMOS_SUCCESS, offset]
        """
        # check if entry config is associated with current track
        if not entry_config.track_name == self.track_name:
            raise SMOS_exceptions.SMOSTrackMismatch(f"Current track is {self.track_name}, while "
                                                    f"input entry_config is associated with track"
                                                    f" {entry_config.track_name}.")

        # calculate result
        if entry_config.mapped_block_idx >= self.max_capacity:
            raise SMOS_exceptions.SMOSMappingError(f"Entry mapped to {entry_config.mapped_block_idx},"
                                                   f"which is out of range ([0, {self.max_capacity - 1}]).")
        else:
            return SMOS_SUCCESS, entry_config.mapped_block_idx * self.block_size


def get_data_track(track_name, shm_name, block_size, max_capacity):
    """
    Returns a data track. This function adds a random tail to shared memory name so that
    there two objects with same name can exist simultaneously in SMOS.

    :param track_name: name of new track
    :param shm_name: name of shared memory space used in new track
    :param block_size: size of each block in new track
    :param max_capacity: max capacity of new track
    :return: a new DataTrack object
    """

    while True:
        try:
            track = DataTrack(track_name=track_name, shm_name=f"{shm_name}_{random.randint(0, SMOS_MAX)}",
                              block_size=block_size, max_capacity=max_capacity)
            return track
        except FileExistsError:
            pass



"""
2021.12.04 Yixuan Mei
This file contains everything that needs to be imported for SMOS.
"""

import SMOS_exceptions
import SMOS_utils

from SMOS_constants import SMOS_SUCCESS, SMOS_FAIL, SMOS_PERMISSION_DENIED
from SMOS_utils import ConnectionDescriptor
from SMOS_server import Server
from SMOS_client import Client, serialize, deserialize



"""
2021.12.01 Yixuan Mei
This file contains class SharedMemoryObjectStore, which is core of SMOS.
"""

import SMOS_exceptions
import SMOS_utils as utils
from SMOS_constants import SMOS_FAIL, SMOS_SUCCESS
from SMOS_shared_memory_object import SharedMemoryObject


class SharedMemoryObjectStore:

    def __init__(self):
        """
        SharedMemoryObjectStore is core to SMOS. It manages all SharedMemoryObjects
        and respond to queries. This class should be initialized using SMOSServer and
        accessed from a SMOSClient.
        """
        self.object_dict = {}
        self.global_lock = utils.RWLock()

    # object manipulation
    def create(self, name, max_capacity, track_count, block_size_list, track_name_list=None):
        """
        Create a new SharedMemoryObject with given parameters in SMOS.

        :exception SMOS_exceptions.SMOSObjectExistError: if object with same name already exists

        :param name: name of the object
        :param max_capacity: maximum number of objects that can be stored in the new SharedMemoryObject
        :param track_count: number of tracks in the new SharedMemoryObject
        :param block_size_list: block size of each track
        :param track_name_list: (optional) name of each track
        :return: always SMOS_SUCCESS
        """
        # create SharedMemoryObject
        self.global_lock.writer_enter()
        if name in self.object_dict:
            raise SMOS_exceptions.SMOSObjectExistError(f"Object with name {name} already exists.")
        self.object_dict[name] = SharedMemoryObject(name=name, max_capacity=max_capacity, track_count=track_count,
                                                    block_size_list=block_size_list, track_name_list=track_name_list)
        self.global_lock.writer_leave()

        # return
        return SMOS_SUCCESS

    def remove(self, name):
        """
        Remove SharedMemoryObject specified by name from SMOS. Note that this is potentially
        destructive since all pending accesses to the object will raise FileNotFound error since
        shared memory space is freed.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if object to be deleted does not exist

        :param name: name of object to be removed
        :return: always SMOSSuccess
        """
        # delete SharedMemoryObject
        self.global_lock.writer_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        shm_object = self.object_dict[name]
        shm_object.stop()
        del shm_object
        del self.object_dict[name]
        self.global_lock.writer_leave()

        # return
        return SMOS_SUCCESS

    # entry manipulation
    # write
    def allocate_block(self, name, entry_config_list: [utils.EntryConfig]):
        """
        Allocate block for new entry in SharedMemoryObject with given name.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :param entry_config_list: configurations of new entry, one for each track
        :return: [SMOS_SUCCESS, entry_config_list] if successful,
                 [SMOS_FAIL, None] if no free block available
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        status = self.object_dict[name].allocate_block(entry_config_list=entry_config_list)
        self.global_lock.reader_leave()

        # return
        if status == SMOS_SUCCESS:
            return SMOS_SUCCESS, entry_config_list
        else:
            return SMOS_FAIL, None

    def append_entry_config(self, name, entry_config_list: [utils.EntryConfig]):
        """
        Append configurations of a new entry into given SharedMemoryObject.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :param entry_config_list: configurations of new entry, one for each track
        :return: always [SMOS_SUCCESS, index for appended entry]
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        _, entry_idx = self.object_dict[name].append_entry_config(entry_config_list=entry_config_list)
        self.global_lock.reader_leave()

        # return
        return SMOS_SUCCESS, entry_idx

    # read
    def read_entry_config(self, name, idx):
        """
        Read entry configuration at given index form given SharedMemoryObject.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :param idx: index of entry to be read
        :return: [SMOS_SUCCESS, entry_config_list] if successful,
                 [SMOS_FAIL, None] if target entry does not exist
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        status, entry_config_list = self.object_dict[name].read_entry_config(idx=idx)
        self.global_lock.reader_leave()

        # return
        return status, entry_config_list

    def release_read_reference(self, name, idx):
        """
        Release read reference on given entry from given SharedMemoryObject.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :param idx: index of entry to be released
        :return: SMOS_SUCCESS if successful,
                 SMOS_FAIL if target entry does not exist
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        status = self.object_dict[name].release_read_reference(idx=idx)
        self.global_lock.reader_leave()

        # return
        return status

    # delete
    def delete_entry_config(self, name, idx, force_delete=False):
        """
        Delete entry at given index from given SharedMemoryObject. Note that this is lazy
        delete, the actual data in shared memory is not erased.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :param idx: index of entry to be deleted
        :param force_delete: whether to delete the entry when there are still pending readers
        :return: SMOS_SUCCESS if successful,
                 SMOS_FAIL if target entry does not exist,
                 SMOS_PERMISSION_DENIED if permission denied
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        status = self.object_dict[name].delete_entry_config(idx=idx, force_delete=force_delete)
        self.global_lock.reader_leave()

        # return
        return status

    # pop and free
    def pop_entry_config(self, name, force_pop=False):
        """
        Pop an entry from SharedMemoryObject. Note that blocks to which the entry is mapped
        will not be freed in this function (since data in the entry will be used after pop).
        Call free_block_mapping to free the block when data in the entry is no longer useful.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :param force_pop: whether to pop the entry when there are still pending readers
        :return: [SMOS_SUCCESS, entry_config_list] if successful,
                 [SMOS_FAIL, None] if data track empty,
                 [SMOS_PERMISSION_DENIED, None] if permission denied
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        status, entry_config_list = self.object_dict[name].pop_entry_config(force_pop=force_pop)
        self.global_lock.reader_leave()

        # return
        return status, entry_config_list

    def free_block_mapping(self, name, entry_config_list: [utils.EntryConfig]):
        """
        Free blocks associated with a previously popped entry.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :param entry_config_list: configurations of entry to be freed
        :return: always SMOS_SUCCESS
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        self.object_dict[name].free_block_mapping(entry_config_list=entry_config_list)
        self.global_lock.reader_leave()

        # return
        return SMOS_SUCCESS

    # stop
    def stop(self):
        """
        Stop SharedMemoryObjectStore for safe exit

        :return: always SMOS_Success
        """
        # stop all objects
        self.global_lock.writer_enter()
        for shm_object in list(self.object_dict.values()):
            shm_object.stop()
        self.global_lock.writer_leave()

        # release lock
        del self.global_lock

    # utility functions
    def get_block_size_list(self, name):
        """
        Get block size of tracks in given SharedMemoryObject.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :return: always [SMOS_SUCCESS, list of block size]
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        block_size_list = self.object_dict[name].block_size_list
        self.global_lock.reader_leave()

        # return
        return SMOS_SUCCESS, block_size_list

    def get_shm_name_list(self, name):
        """
        Get names of all shared memories associated with given SharedMemoryObject.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :return: always [SMOS_SUCCESS, list of shared memory names]
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        shm_name_list = self.object_dict[name].get_shm_name_list()
        self.global_lock.reader_leave()

        # return
        return SMOS_SUCCESS, shm_name_list

    def get_entry_offset_list(self, name, entry_config_list: [utils.EntryConfig]):
        """
        Get offset of each track for given entry in shared memory space.

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :param entry_config_list: configurations of entry to be queried
        :return: always [SMOS_SUCCESS, offset_list]
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        status, offset_list = self.object_dict[name].get_entry_offset(entry_config_list=entry_config_list)
        self.global_lock.reader_leave()

        # return
        return status, offset_list

    def get_entry_idx_list(self, name):
        """
        Get all entry index in given SharedMemoryObject

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :return: always [SMOS_SUCCESS, entry_idx_list]
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        entry_idx_list = self.object_dict[name].get_entry_idx_list()
        self.global_lock.reader_leave()

        # return
        return SMOS_SUCCESS, entry_idx_list

    def get_track_count(self, name):
        """
        Get track count of target SharedMemoryObject

        :exception SMOS_exceptions.SMOSObjectNotFoundError: if target SharedMemoryObject
                   does not exist

        :param name: name of the SharedMemoryObject
        :return: always [SMOS_SUCCESS, track count]
        """
        # query target SharedMemoryObject
        self.global_lock.reader_enter()
        if name not in self.object_dict:
            raise SMOS_exceptions.SMOSObjectNotFoundError(f"Object with name {name} not found.")
        track_count = self.object_dict[name].track_count
        self.global_lock.reader_leave()

        # return
        return SMOS_SUCCESS, track_count

    def profile(self):
        """
        Profiles SMOS usage.
        """
        # fetch data
        self.global_lock.reader_enter()
        name_list, entry_count_list, capacity_list = [], [], []
        for shm_object in list(self.object_dict.values()):
            name_list.append(shm_object.name)
            entry_count_list.append(shm_object.get_entry_count())
            capacity_list.append(shm_object.max_capacity)
        self.global_lock.reader_leave()

        # print to terminal
        utils.log2terminal(info_type="Info", msg=f"start profiling:")
        for idx in range(len(name_list)):
            utils.log2terminal(info_type="Info", msg=f"{name_list[idx]}: {entry_count_list[idx]}"
                                                     f"/{capacity_list[idx]}")



"""
2021.12.01 Yixuan Mei
This file contains class Client, which should be instantiated in every process that uses SMOS.
"""

import pickle
from multiprocessing import shared_memory

import numpy as np

import SMOS_exceptions
import SMOS_server
import SMOS_utils as utils
from SMOS_constants import SMOS_SUCCESS, SMOS_FAIL
from SMOS_utils import safe_execute


class Client:

    def __init__(self, connection: utils.ConnectionDescriptor):
        """
        Each Client instance connects to one existing server to access SMOS
        object storage services.

        :param connection: a ConnectionDescriptor that specifies address of SMOS server
        """
        self.connection = connection
        self.store = SMOS_server.get_object_store(connection=connection)
        utils.remove_shm_from_resource_tracker()

    # SharedMemoryObject operations
    #   workflow 1:  create_object  ->  [entry operations]  ->  remove_object
    #   workflow 2:  put            ->  (multiple) get
    # These two usages can be combined.
    def create_object(self, name, max_capacity, track_count, block_size, track_name=None):
        """
        Create a new SharedMemoryObject with given parameters in SMOS.

        :param name: name of the object
        :param max_capacity: maximum number of objects that can be stored in the new SharedMemoryObject
        :param track_count: number of tracks in the new SharedMemoryObject
        :param block_size: block size of each track, a list if there are multiple tracks
        :param track_name: (optional) name of each track, a list if there are multiple tracks
        :return: always SMOS_SUCCESS
        """
        # ensure that block_size and track_name are lists
        if not type(block_size) == list:
            block_size = [block_size]
        if track_name is not None and not type(track_name) == list:
            track_name = [track_name]

        # create object
        status = safe_execute(target=self.store.create, args=(name, max_capacity, track_count,
                                                              block_size, track_name,))
        return status

    def remove_object(self, name):
        """
        Remove SharedMemoryObject specified by name from SMOS. Note that this is potentially
        destructive since all pending accesses to the object will raise FileNotFound error since
        shared memory space is freed.

        :param name: name of object to be removed
        :return: always SMOSSuccess
        """
        # create object
        status = safe_execute(target=self.store.remove, args=(name,))
        return status

    def put(self, name, data, as_list=False, redundancy=0):
        """
        Create a SharedMemoryObject with name,  and put data into this SharedMemoryObject.

        :exception SMOS_exceptions.SMOSDimensionMismatch: if multiple entries have different
                   number of tracks

        :param name: name of share memory object
        :param data: data to be put into SMOS
        :param as_list: If as_list = True and data is a list, then each element of the list
               will be stored in a single entry in the SharedMemoryObject. In other cases,
               data will be stored as a whole in one entry and the SharedMemoryObject will
               have only one entry.
        :param redundancy: Require redundancy number of extra free entries in new
               SharedMemoryObject. These extra entries can be used by entry operations.
        :return: always SMOS_SUCCESS
        """
        # format data and check integrity
        if not as_list or not type(data) == list:
            data = [data]
        track_count_list = []
        for entry_idx in range(len(data)):
            if not type(data[entry_idx]) == list:
                data[entry_idx] = [data[entry_idx]]
            track_count_list.append(len(data[entry_idx]))
        if not len(set(track_count_list)) == 1:
            raise SMOS_exceptions.SMOSDimensionMismatch("Multiple entries have different number"
                                                        "of tracks.")
        entry_count = len(data)
        track_count = track_count_list[0]

        # serialize data and calculate size
        serialized_data, data_size_list, data_is_numpy_list = [], [], []
        for data_entry in data:
            serialized_entry = []
            entry_size_list = []
            is_numpy_list = []
            for element in data_entry:
                if type(element) == np.ndarray:
                    serialized_entry.append(element)
                    entry_size_list.append(element.nbytes)
                    is_numpy_list.append(True)
                else:
                    serialized_stream = utils.serialize(element)
                    serialized_entry.append(serialized_stream)
                    entry_size_list.append(len(serialized_stream))
                    is_numpy_list.append(False)
            serialized_data.append(serialized_entry)
            data_size_list.append(entry_size_list)
            data_is_numpy_list.append(is_numpy_list)

        # calculate max size and create SharedMemoryObject
        block_size_list = [max(np.array(data_size_list)[:, i]) for i in range(track_count)]
        self.create_object(name=name, max_capacity=entry_count + redundancy, track_count=track_count,
                           block_size=block_size_list)

        # create entry, write into shared memory and commit
        for entry_idx in range(entry_count):
            # create entry
            serialized_entry = serialized_data[entry_idx]
            is_numpy_list = data_is_numpy_list[entry_idx]
            dtype_list, shape_list = [], []
            for track_idx in range(track_count):
                if is_numpy_list[track_idx]:
                    dtype_list.append(serialized_entry[track_idx].dtype)
                    shape_list.append(serialized_entry[track_idx].shape)
                else:
                    dtype_list.append(None)
                    shape_list.append(None)
            _, object_handle = self.create_entry(name=name, dtype=dtype_list, shape=shape_list,
                                                 is_numpy=is_numpy_list)

            # write into shared memory
            _, buffer_list = self.open_shm(object_handle=object_handle)
            for track_idx in range(track_count):
                if is_numpy_list[track_idx]:
                    buffer_list[track_idx][:] = serialized_entry[track_idx][:]
                else:
                    stream_length = len(serialized_entry[track_idx])
                    buffer_list[track_idx][0:stream_length] = serialized_entry[track_idx]

            # commit
            self.commit_entry(object_handle=object_handle)

        # return
        return SMOS_SUCCESS

    def get(self, name, entry_idx_list=None):
        """
        Get a SharedMemoryObject from shared memory. The object will be reconstructed
        to what it was before being passed into SMOS. Note that this function always
        copies the data from shared memory.

        :param name: name of object to get
        :param entry_idx_list: if entry_idx_list is None, all entries in this
               SharedMemoryObject will be returned
        :return: reconstructed_object
        """
        # determine query range
        if entry_idx_list is None:
            _, entry_idx_list = self.store.get_entry_idx_list(name=name)

        # read and reconstruct data from shared memory
        return_list = []
        for entry_idx in entry_idx_list:
            # read reconstructed object
            status, handle, obj = self.read_from_object(name=name, entry_idx=entry_idx)
            if status == SMOS_FAIL:
                return_list.append(None)
                continue

            # further reconstruction
            if type(obj) == np.ndarray:
                obj = obj.copy()
            if type(obj) == list:
                for track_idx in range(len(obj)):
                    if type(obj[track_idx]) == np.ndarray:
                        obj[track_idx] = obj[track_idx].copy()

            # append to return list and release handle
            return_list.append(obj)
            self.release_entry(object_handle=handle)

        # return
        if len(return_list) == 1:
            return return_list[0]
        else:
            return return_list

    # entry operations
    # fine-grained operations (zero copy)
    #    create procedure:  create_entry  ->  open_shm  ->  commit_entry
    #    r/w procedure:     open_entry    ->  open_shm  ->  release_entry
    #    delete procedure:  delete_entry
    def create_entry(self, name, dtype, shape, is_numpy):
        """
        Create a new entry in given SharedMemoryObject.

        :exception SMOS_exceptions.SMOSDimensionMismatch: if length of input is
                   different of number of tracks

        :param name: name of target SharedMemoryObject
        :param dtype: dtype for each track (None if not numpy), a list if there are multiple tracks
        :param shape: shape for each track (None if not numpy), a list if there are multiple tracks
        :param is_numpy: if track stores numpy array, a list if there are multiple tracks
        :return: [SMOS_SUCCESS, ObjectHandle] if successful
                 [SMOS_FAIL, None] if no free space available in target object
        """
        # ensure that inputs are lists
        if not type(dtype) == list:
            dtype_list = [dtype]
        else:
            dtype_list = dtype
        if not type(shape) == list:
            shape_list = [shape]
        else:
            shape_list = shape
        if not type(is_numpy) == list:
            is_numpy_list = [is_numpy]
        else:
            is_numpy_list = is_numpy

        # check input integrity
        _, track_count = safe_execute(target=self.store.get_track_count, args=(name,))
        if not len(dtype_list) == len(shape_list) == track_count:
            raise SMOS_exceptions.SMOSDimensionMismatch(f"There are {track_count} tracks, but only "
                                                        f"{len(dtype_list)} input dtypes and {len(shape_list)}"
                                                        f"input shapes.")

        # prepare entry config
        entry_config_list = []
        for track_idx in range(track_count):
            entry_config = utils.EntryConfig(dtype=dtype_list[track_idx], shape=shape_list[track_idx],
                                             is_numpy=is_numpy_list[track_idx])
            entry_config_list.append(entry_config)

        # allocate block
        status, entry_config_list = safe_execute(target=self.store.allocate_block,
                                                 args=(name, entry_config_list,))

        # build object handle and return
        if status == SMOS_SUCCESS:
            object_handle = utils.ObjectHandle()
            object_handle.name = name
            object_handle.track_count = track_count
            object_handle.entry_config_list = entry_config_list
            return SMOS_SUCCESS, object_handle
        else:
            return SMOS_FAIL, None

    def open_entry(self, name, entry_idx):
        """
        Open entry specified by entry_idx from target SharedMemoryObject.

        :param name: name of the SharedMemoryObject
        :param entry_idx: index of entry to be read
        :return: [SMOS_SUCCESS, ObjectHandle] if successful
                 [SMOS_FAIL, None] if index out of range
        """
        # get entry config
        status, entry_config_list = safe_execute(target=self.store.read_entry_config,
                                                 args=(name, entry_idx,))

        # build object handle
        if status == SMOS_SUCCESS:
            object_handle = utils.ObjectHandle()
            object_handle.name = name
            object_handle.entry_idx = entry_idx
            object_handle.track_count = len(entry_config_list)
            object_handle.entry_config_list = entry_config_list
            return SMOS_SUCCESS, object_handle
        else:
            return SMOS_FAIL, None

    def open_shm(self, object_handle: utils.ObjectHandle):
        """
        Open data stored in target entry from shared memory space. For tracks that
        store numpy array, the return value is a shared memory backed numpy array.
        For tracks that store other customized types of data, return value is raw
        buffer (shm.buf).

        :param object_handle: handle of target entry
        :return: always [SMOS_SUCCESS, a list of shm.buf / shm backed numpy array]
        """
        # get offset and shared memory name
        _, offset_list = safe_execute(target=self.store.get_entry_offset_list,
                                      args=(object_handle.name, object_handle.entry_config_list,))
        _, block_size_list = safe_execute(target=self.store.get_block_size_list,
                                          args=(object_handle.name,))
        _, shm_name_list = safe_execute(target=self.store.get_shm_name_list,
                                        args=(object_handle.name,))

        # open shared memory and get data
        return_list = []
        for track_idx in range(object_handle.track_count):
            # open shm
            shm = shared_memory.SharedMemory(name=shm_name_list[track_idx])
            object_handle.shm_list.append(shm)

            # open data in different formats
            entry_config = object_handle.entry_config_list[track_idx]
            if entry_config.is_numpy:
                mm_array = np.ndarray(shape=entry_config.shape, dtype=entry_config.dtype,
                                      buffer=shm.buf, offset=offset_list[track_idx])
                return_list.append(mm_array)
            else:
                buffer = shm.buf[offset_list[track_idx]: offset_list[track_idx] + block_size_list[track_idx]]
                object_handle.buf_list.append(buffer)
                return_list.append(buffer)

        # return
        return SMOS_SUCCESS, return_list

    def commit_entry(self, object_handle: utils.ObjectHandle):
        """
        Commit the entry to SMOS. After this operation the entry will be visible to
        all other processes. Note that after commit, the object_handle will be destroyed.

        :param object_handle: handle of target entry
        :return: always [SMOS_SUCCESS, index of newly committed entry]
        """
        # commit entry config
        _, entry_idx = safe_execute(target=self.store.append_entry_config,
                                    args=(object_handle.name, object_handle.entry_config_list,))

        # clean up
        for buffer in object_handle.buf_list:
            buffer.release()
        for shm in object_handle.shm_list:
            shm.close()

        # return
        return SMOS_SUCCESS, entry_idx

    def release_entry(self, object_handle: utils.ObjectHandle):
        """
        Release read reference to target entry when reading is finished. Note that
        after this operation, object_handle is destroyed.

        :param object_handle: entry to be released
        :return: SMOS_SUCCESS if successful,
                 SMOS_FAIL if target entry does not exist (has been deleted)
        """
        # release read reference
        status = safe_execute(target=self.store.release_read_reference,
                              args=(object_handle.name, object_handle.entry_idx,))

        # clean up
        if status == SMOS_SUCCESS:
            for buffer in object_handle.buf_list:
                buffer.release()
            for shm in object_handle.shm_list:
                shm.close()

        # return
        return status

    def delete_entry(self, name, entry_idx, force_delete=False):
        """
        Delete target entry from SharedMemoryObject specified by name. Note that this is lazy
        delete, the actual data in shared memory is not erased.

        :param name: name of the SharedMemoryObject
        :param entry_idx: index of entry to be deleted
        :param force_delete: whether to delete the entry when there are still pending readers
        :return: SMOS_SUCCESS if successful,
                 SMOS_FAIL if target entry does not exist,
                 SMOS_PERMISSION_DENIED if permission denied
        """
        # delete entry
        status = safe_execute(target=self.store.delete_entry_config,
                              args=(name, entry_idx, force_delete,))
        return status

    # coarse-grained operations (queue API)
    #   write process: push_to_object
    #   read process:  pop_from_object  ->  free_handle
    #                  read_from_object ->  release_entry
    def pop_from_object(self, name, force_pop=False):
        """
        Pop an entry from target object. This function reconstructs data to what it
        was before being passed into SMOS.

        :param name: name of the SharedMemoryObject
        :param force_pop: whether to pop an entry when there are still pending readers
        :return: [SMOS_SUCCESS, object_handle, reconstructed_object] if successful
                 [SMOS_FAIL, None, None] if target SharedMemoryObject is empty
                 [SMOS_PERMISSION_DENIED, None, None] if permission denied
        """
        # pop entry config
        status, entry_config_list = safe_execute(target=self.store.pop_entry_config,
                                                 args=(name, force_pop,))

        # check if we successfully get entry config
        if not status == SMOS_SUCCESS:
            return status, None, None

        # get shared memory info
        _, offset_list = safe_execute(target=self.store.get_entry_offset_list,
                                      args=(name, entry_config_list,))
        _, block_size_list = safe_execute(target=self.store.get_block_size_list, args=(name,))
        _, shm_name_list = safe_execute(target=self.store.get_shm_name_list, args=(name,))

        # build object handle (partial build)
        object_handle = utils.ObjectHandle()
        object_handle.name = name
        object_handle.entry_config_list = entry_config_list

        # reconstruct object
        reconstructed_object = []
        for track_idx in range(len(entry_config_list)):
            shm = shared_memory.SharedMemory(name=shm_name_list[track_idx])
            entry_config = entry_config_list[track_idx]
            if entry_config.is_numpy:
                mm_array = np.ndarray(shape=entry_config.shape, dtype=entry_config.dtype,
                                      buffer=shm.buf, offset=offset_list[track_idx])
                object_handle.shm_list.append(shm)
                reconstructed_object.append(mm_array)
            else:
                buffer = shm.buf[offset_list[track_idx]: offset_list[track_idx] + block_size_list[track_idx]]
                deserialized_object = utils.deserialize(data_stream=buffer)
                buffer.release()
                reconstructed_object.append(deserialized_object)

        # return
        if len(reconstructed_object) == 1:
            return SMOS_SUCCESS, object_handle, reconstructed_object[0]
        else:
            return SMOS_SUCCESS, object_handle, reconstructed_object

    def read_from_object(self, name, entry_idx):
        """
        Read an entry from target SharedMemoryObject. This function reconstructs data
        to what it was before being passed into SMOS.

        :param name: name of the SharedMemoryObject
        :param entry_idx: index of entry to be read
        :return: [SMOS_SUCCESS, object_handle, reconstructed_object] if successful
                 [SMOS_FAIL, None, None] if entry does not exist in target SharedMemoryObject
        """
        # pop entry config
        status, entry_config_list = safe_execute(target=self.store.read_entry_config,
                                                 args=(name, entry_idx,))

        # check if we successfully get entry config
        if not status == SMOS_SUCCESS:
            return status, None, None

        # get shared memory info
        _, offset_list = safe_execute(target=self.store.get_entry_offset_list,
                                      args=(name, entry_config_list,))
        _, block_size_list = safe_execute(target=self.store.get_block_size_list, args=(name,))
        _, shm_name_list = safe_execute(target=self.store.get_shm_name_list, args=(name,))

        # build object handle (partial build)
        object_handle = utils.ObjectHandle()
        object_handle.name = name
        object_handle.entry_idx = entry_idx
        object_handle.entry_config_list = entry_config_list

        # reconstruct object
        reconstructed_object = []
        for track_idx in range(len(entry_config_list)):
            shm = shared_memory.SharedMemory(name=shm_name_list[track_idx])
            entry_config = entry_config_list[track_idx]
            if entry_config.is_numpy:
                mm_array = np.ndarray(shape=entry_config.shape, dtype=entry_config.dtype,
                                      buffer=shm.buf, offset=offset_list[track_idx])
                object_handle.shm_list.append(shm)
                reconstructed_object.append(mm_array)
            else:
                buffer = shm.buf[offset_list[track_idx]: offset_list[track_idx] + block_size_list[track_idx]]
                deserialized_object = utils.deserialize(data_stream=buffer)
                buffer.release()
                reconstructed_object.append(deserialized_object)

        # return
        if len(reconstructed_object) == 1:
            return SMOS_SUCCESS, object_handle, reconstructed_object[0]
        else:
            return SMOS_SUCCESS, object_handle, reconstructed_object

    def free_handle(self, object_handle: utils.ObjectHandle):
        """
        Free an object_handle returned by a previous pop_from_object. This operation will
        release all SMOS resources related with the handle. reconstructed_object's shared
        memory arrays can no longer be accessed once this function is called.

        :param object_handle: object handle to free
        :return: always SMOS_SUCCESS
        """
        # free block mapping of underlying entry
        _ = safe_execute(target=self.store.free_block_mapping,
                         args=(object_handle.name, object_handle.entry_config_list))

        # release resources
        for shm in object_handle.shm_list:
            shm.close()

        # return
        return SMOS_SUCCESS

    def push_to_object(self, name, data):
        """
        Push data to target SharedMemoryObject. Note that if data is a list, then each element
        in the list will be assigned to one track in SharedMemoryObject.

        :exception SMOS_exceptions.SMOSDimensionMismatch: if the number of elements in data is
                   different from track_count of target SharedMemoryObject

        :param name: name of the SharedMemoryObject
        :param data: data to be pushed
        :return: [SMOS_SUCCESS, entry_idx] if successful
                 [SMOS_FAIL, None] if no free space available in target object
        """
        # ensure that data is a list
        if not type(data) == list:
            data = [data]

        # check input dimension
        _, track_count = safe_execute(target=self.store.get_track_count, args=(name,))
        if not len(data) == track_count:
            raise SMOS_exceptions.SMOSDimensionMismatch(f"There are {track_count} tracks, but input data"
                                                        f"has length {len(data)}.")

        # compute input data configuration
        dtype_list, shape_list, is_numpy_list = [], [], []
        for data_element in data:
            if type(data_element) == np.ndarray:
                dtype_list.append(data_element.dtype)
                shape_list.append(data_element.shape)
                is_numpy_list.append(True)
            else:
                dtype_list.append(None)
                shape_list.append(None)
                is_numpy_list.append(False)

        # create new entry
        status, object_handle = self.create_entry(name=name, dtype=dtype_list, shape=shape_list,
                                                  is_numpy=is_numpy_list)
        if status == SMOS_FAIL:
            return SMOS_FAIL, None

        # serialize input data and copy into shared memory
        _, buffer_list = self.open_shm(object_handle=object_handle)
        for track_idx in range(track_count):
            if is_numpy_list[track_idx]:
                buffer_list[track_idx][:] = data[track_idx][:]
            else:
                utils.serialize(obj=data[track_idx], file=buffer_list[track_idx].obj)

        # commit entry and return
        _, entry_idx = self.commit_entry(object_handle=object_handle)
        return SMOS_SUCCESS, entry_idx


# serialize and deserialize (one-copy)
def serialize(obj, buf):
    """
    Serialize target object into buffer. Note that this should only be
    used for non-numpy objects, since numpy arrays have more efficient
    zero-copy ser/des protocol.

    :param obj: object to be serialized
    :param buf: a memoryview object, e.g. the one returned by open_shm
           for non-numpy objects.
    """
    # check input
    if not type(buf) == memoryview:
        raise SMOS_exceptions.SMOSInputTypeError(f"memoryview expected for buf, get {type(buf)}")

    # serialize
    pickle.dump(obj=obj, file=buf.obj, protocol=pickle.HIGHEST_PROTOCOL)


def deserialize(buf):
    """
    Deserialize target object from buffer. Note that this should only be
    used for non-numpy objects, since numpy arrays have more efficient
    zero-copy ser/des protocol.

    :param buf: buffer that stores the data stream
    :return: deserialized_object
    """
    deserialized_object = pickle.loads(data=buf)
    return deserialized_object



"""
2021.11.24 Yixuan Mei
This file contains utility functions for Shared Memory Object Store
"""

import multiprocessing
import os
import pickle
import socket
from multiprocessing import resource_tracker

import numpy as np

import SMOS_exceptions


class EntryConfig:

    def __init__(self, dtype, shape, is_numpy, track_name=None, mapped_block_idx=-1):
        """
        Each entry config represents configuration of an entry in one track

        :param dtype: If current entry represents a numpy array, dtype is element type
                      of this numpy array. Otherwise, dtype is type of the object.
        :param shape: If current entry represents a numpy array, shape is the shape of
                      this numpy array. Otherwise, shape is None.
        :param is_numpy: Whether this entry represent a numpy array.
        :param track_name: name of the track associated with this entry
        :param mapped_block_idx: Index of the block in which current entry is stored.
        """
        # data configuration
        self.dtype = dtype
        self.shape = shape
        self.is_numpy = is_numpy

        # management
        self.track_name = track_name
        self.mapped_block_idx = mapped_block_idx
        self.pending_readers = 0


class RWLock:

    def __init__(self):
        """
        This class implements a fair RW lock between multiple readers and writers.
        """
        self.reader_counter_lock = multiprocessing.Lock()
        self.writer_lock = multiprocessing.Lock()
        self.readwrite_lock = multiprocessing.Lock()
        self.reader_count = 0

    def reader_enter(self):
        self.writer_lock.acquire()
        self.reader_counter_lock.acquire()
        if self.reader_count == 0:
            self.readwrite_lock.acquire()
        self.reader_count += 1
        self.reader_counter_lock.release()
        self.writer_lock.release()

    def reader_leave(self):
        self.reader_counter_lock.acquire()
        self.reader_count -= 1
        if self.reader_count == 0:
            self.readwrite_lock.release()
        self.reader_counter_lock.release()

    def writer_enter(self):
        self.writer_lock.acquire()
        self.readwrite_lock.acquire()

    def writer_leave(self):
        self.readwrite_lock.release()
        self.writer_lock.release()


class ConnectionDescriptor:

    def __init__(self, ip, port, authkey):
        """
        Describes the location of a python multiprocessing manager

        :param ip: ip address, could be string 'localhost'
        :param port: 0 - 65535
        :param authkey: a bytes string

        """
        self.ip = ip
        self.port = port
        self.authkey = authkey


class ObjectHandle:

    def __init__(self):
        """
        Each ObjectHandle represents an entry of SharedMemoryObject in remote
        SMOS server.
        """
        # basic information
        self.name = None
        self.entry_idx = None
        self.track_count = None

        # management
        self.entry_config_list = None
        self.shm_list = []
        self.buf_list = []  # this only contains shm.buf, used for clean up


def get_local_free_port(num, low, high):
    """
    Get num free ports between low and high if possible.

    :exception object_store_exceptions.SMOSPortBusy:  If not enough free ports are available

    :param num: number of free ports required
    :param low: lower bound of port idx
    :param high: upper bound of port idx
    :return: free port list
    """
    port_list = []
    for port in range(low, high):
        # check if port is in use
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            is_busy = s.connect_ex(('localhost', port)) == 0
        if not is_busy:
            port_list.append(port)
        if len(port_list) == num:
            return port_list
    raise SMOS_exceptions.SMOSPortBusy(f"Out of free ports (required {num}, available {len(port_list)}).")


def safe_execute(target, args=()):
    """
    Automatically retry when python multiprocessing manager fails due to
    port number drainage. Wrap this outside every remote call to server
    functions.

    :param target: the function to run
    :param args: arguments of target function
    :return: what the function returns
    """
    fail_counter = 0
    while True:
        try:
            return target(*args)
        except TypeError:
            if fail_counter > 10:
                raise TypeError
            else:
                fail_counter += 1
                log2terminal(info_type="Warning", msg=f"Proxy fails {fail_counter} times. Retrying...")


def remove_shm_from_resource_tracker():
    """Monkey-patch multiprocessing.resource_tracker so SharedMemory won't be tracked

    More details at: https://bugs.python.org/issue38119
    """

    def fix_register(name, rtype):
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.register(self, name, rtype)

    resource_tracker.register = fix_register

    def fix_unregister(name, rtype):
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.unregister(self, name, rtype)

    resource_tracker.unregister = fix_unregister

    if "shared_memory" in resource_tracker._CLEANUP_FUNCS:
        del resource_tracker._CLEANUP_FUNCS["shared_memory"]


def log2terminal(info_type, msg, worker_type=""):
    """
    log message to terminal in uniform format

    :param info_type: Usually Error / Warning / Info
    :param msg: message body
    :param worker_type: (optional) Identifies who send this message.
    """
    print(f"[(pid={os.getpid()}){worker_type}] {info_type}: {msg}", flush=True)


def serialize_numpy(numpy_array):
    """
    Serialize a numpy array for one track. This is the recommended way of putting a
     numpy array into shared memory object store.

    :exception object_store_exceptions.SMOSInputTypeError: If input is not a numpy array

    :param numpy_array: numpy array to be serialized
    :return: entry_config, numpy_array
    """
    # check if input is a numpy array
    if not type(numpy_array) == np.ndarray:
        raise SMOS_exceptions.SMOSInputTypeError(f"Input not numpy array.")

    # construct entry config for each array
    entry_config = EntryConfig(dtype=numpy_array.dtype, shape=numpy_array.shape, is_numpy=True)
    return entry_config, numpy_array


def serialize_numpy_list(numpy_list):
    """
    Serialize a list of numpy arrays for multiple tracks. This is the recommended
    way of putting a list of numpy arrays into shared memory object store.

    :exception object_store_exceptions.SMOSInputTypeError:  If input is not a list of
                                                            numpy arrays

    :param numpy_list: a list of numpy to be serialized
    :return: entry_config_list, numpy_list
    """
    # check if input is a list of numpy arrays
    if not type(numpy_list) == list:
        raise SMOS_exceptions.SMOSInputTypeError(f"Input not list.")
    if len(numpy_list) == 0:
        return [], []
    if not type(numpy_list[0]) == np.ndarray:
        raise SMOS_exceptions.SMOSInputTypeError(f"Expected numpy list, got list of"
                                                 f" {type(numpy_list[0])}.")

    # construct entry config for each array
    entry_config_list = []
    for np_array in numpy_list:
        entry_config = EntryConfig(dtype=np_array.dtype, shape=np_array.shape, is_numpy=True)
        entry_config_list.append(entry_config)
    return entry_config_list, numpy_list


def deserialize_numpy(entry_config: EntryConfig, offset, buffer):
    """
    Deserialize a numpy array from buffer (zero copy).

    :exception SMOS_exceptions.SMOSInputTypeError: if entry does not contain data of
               a numpy array

    :param entry_config: configuration of target entry
    :param offset: offset in buffer
    :param buffer: buffer that stores the data
    :return: deserialized numpy array
    """
    # check if input is numpy
    if not entry_config.is_numpy:
        raise SMOS_exceptions.SMOSInputTypeError("Current entry is not numpy.")

    # deserialize
    deserialized_array = np.ndarray(shape=entry_config.shape, dtype=entry_config.dtype,
                                    buffer=buffer, offset=offset)
    return deserialized_array


def deserialize_numpy_list(entry_config_list: [EntryConfig], offset_list, buffer_list):
    """
    Deserialize a list of numpy arrays from buffers (zero copy).

    :param entry_config_list: configuration of target entries
    :param offset_list: offset in buffer
    :param buffer_list: buffer that stores the data
    :return: a list of deserialized numpy arrays
    """
    # deserialize
    deserialized_array_list = []
    for idx in range(len(entry_config_list)):
        deserialized_array = deserialize_numpy(entry_config=entry_config_list[idx],
                                               offset=offset_list[idx], buffer=buffer_list[idx])
        deserialized_array_list.append(deserialized_array)

    # return
    return deserialized_array_list


def serialize(obj, file=None):
    """
    This is a general serializer based on pickle. Note that (list of) numpy arrays
    should use serialize_numpy(_list) instead for better performance.

    :param obj: object to be serialized
    :param file: file to write the stream into
    :return: data_stream if file is None
             no return if file is not None
    """
    if file is None:
        data_stream = pickle.dumps(obj=obj, protocol=pickle.HIGHEST_PROTOCOL)
        return data_stream
    else:
        pickle.dump(obj=obj, file=file, protocol=pickle.HIGHEST_PROTOCOL)


def serialize_list(obj_list):
    """
    This is a general serializer based on pickle. Note that (list of) numpy arrays
    should use serialize_numpy(_list) instead for better performance.

    :exception object_store_exceptions.SMOSInputTypeError: If input is not a list.

    :param obj_list: list of objects to be serialized
    :return: data_stream_list
    """
    # check if input is a list
    if not type(obj_list) == list:
        raise SMOS_exceptions.SMOSInputTypeError("Input not list")

    # serialize
    data_stream_list = []
    for obj in obj_list:
        data_stream = serialize(obj)
        data_stream_list.append(data_stream)
    return data_stream_list


def deserialize(data_stream):
    """
    This is a general deserializer based on pickle. Note that data_stream must be
    the result of previous calls to serialize.

    :param data_stream: data stream to be deserialized
    :return: deserialized_object
    """
    deserialized_object = pickle.loads(data=data_stream)
    return deserialized_object


def deserialize_list(data_stream_list):
    """
    This is a general deserializer based on pickle. Note that data_stream_list must be
    the result of previous calls to serialize_list.

    :exception object_store_exceptions.SMOSInputTypeError: If input is not a list.

    :param data_stream_list: list of data streams to be deserialized
    :return: deserialized_object_list
    """
    # check if input is a list
    if not type(data_stream_list) == list:
        raise SMOS_exceptions.SMOSInputTypeError("Input not list")

    # deserialize
    deserialized_object_list = []
    for data_stream in data_stream_list:
        deserialized_object_list.append(deserialize(data_stream))
    return deserialized_object_list



"""
2021.11.25 Yixuan Mei
This file contains constants used in Shared Memory Object Store
"""

import sys

# return values
SMOS_FAIL = -1
SMOS_SUCCESS = 0
SMOS_PERMISSION_DENIED = 1

# math
SMOS_MAX = sys.maxsize



import setuptools

# usage
# python setup.py sdist bdist_wheel
# pip install -e .

# readme
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="SMOS_antony",  # This is the name of the package
    version="1.0.0",  # The initial release version
    author="Yixuan Mei",  # Full name of the author
    author_email="meiyx19@mails.tsinghua.edu.cn",
    url="https://github.com/AntonyMei/SharedMemoryObjectStore",
    description="A fast shared memory object store.",
    long_description=long_description,  # Long description read from the the readme file
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),  # List of all python modules to be installed
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],  # Information to filter the project on PyPi website
    python_requires='>=3.8',  # Minimum version requirement of the package
    py_modules=["SMOS", "SMOS_client", "SMOS_constants", "SMOS_data_track", "SMOS_exceptions", "SMOS_server",
                "SMOS_shared_memory_object", "SMOS_shared_memory_object_store", "SMOS_utils"],
    # Name of the python package
    package_dir={'': 'src'},  # Directory of the source code of the package
    install_requires=["numpy"]  # Install other dependencies if any
)



"""
2021.11.29 Yixuan Mei
This file contains class SharedMemoryObject, which is basic component of SMOS.
"""

import SMOS_exceptions
import SMOS_utils as utils
from SMOS_constants import SMOS_FAIL, SMOS_SUCCESS
from SMOS_data_track import get_data_track


class SharedMemoryObject:

    def __init__(self, name, max_capacity, track_count, block_size_list, track_name_list=None):
        """
        SharedMemoryObject is a basic component of SMOS. Each object stored in SMOS is represented
        by a SharedMemoryObject. Such objects can have multiple DataTracks for its data components.
        In order to avoid concurrency issues, shared memory object uses RWLock to synchronize accesses
        from different processes.

        :exception SMOS_exceptions.SMOSDimensionMismatch: if length of block_size_list does not
                   equal to track_count

        :param name: name of this SharedMemoryObject
        :param max_capacity: maximum number of objects that can be stored in this SharedMemoryObject
        :param track_count: number of tracks in this SharedMemoryObject
        :param block_size_list: block size of each track
        :param track_name_list: (optional) name of each track
        """
        # save parameters
        self.name = name
        self.track_count = track_count
        self.max_capacity = max_capacity

        # save block sizes
        if not len(block_size_list) == track_count:
            raise SMOS_exceptions.SMOSDimensionMismatch(f"There are {track_count} tracks, but only"
                                                        f"{len(block_size_list)} items in block size list.")
        self.block_size_list = block_size_list

        # save track names
        if track_name_list is not None and len(track_name_list) == track_count:
            self.track_name_list = track_name_list
        else:
            self.track_name_list = list(range(track_count))

        # prepare tracks
        self.track_list = []
        for track_idx in range(track_count):
            track = get_data_track(track_name=f"{name}_{self.track_name_list[track_idx]}",
                                   shm_name=f"{name}_{self.track_name_list[track_idx]}",
                                   block_size=self.block_size_list[track_idx], max_capacity=self.max_capacity)
            self.track_list.append(track)

        # save shared memory names
        self.shm_name_list = [track.shm_name for track in self.track_list]

        # prepare lock
        self.lock = utils.RWLock()

    # write
    def allocate_block(self, entry_config_list: [utils.EntryConfig]):
        """
        Allocate a free block for a new entry and write into entry config.

        :exception SMOS_exceptions.SMOSDimensionMismatch: if length of input is
                   different of number of tracks
        :exception SMOS_exceptions.SMOSTrackUnaligned: if some tracks are able to
                   allocate a new block while others are not

        :param entry_config_list: configurations of new entry, one for each track
        :return: SMOS_SUCCESS if successful,
                 SMOS_FAIL if no free block available
        """
        # check input shape
        if not len(entry_config_list) == self.track_count:
            raise SMOS_exceptions.SMOSDimensionMismatch(f"There are {self.track_count} tracks, but only"
                                                        f"{len(entry_config_list)} items in input.")

        # allocate block
        status_list = []
        for track_idx in range(self.track_count):
            status = self.track_list[track_idx].allocate_block(entry_config=entry_config_list[track_idx])
            status_list.append(status)

        # return
        if not len(set(status_list)) == 1:
            raise SMOS_exceptions.SMOSTrackUnaligned("Track unaligned.")
        return status_list[0]

    def append_entry_config(self, entry_config_list: [utils.EntryConfig]):
        """
        Append configurations of new entry into SharedMemoryObject's tracks, one
        configuration for each track.

        :exception SMOS_exceptions.SMOSDimensionMismatch: if length of input is
                   different of number of tracks
        :exception SMOS_exceptions.SMOSTrackUnaligned: if different tracks return
                   different index for newly appended entry

        :param entry_config_list: configurations of new entry, one for each track
        :return: always [SMOS_SUCCESS, index of appended entry config]
        """
        # check input shape
        if not len(entry_config_list) == self.track_count:
            raise SMOS_exceptions.SMOSDimensionMismatch(f"There are {self.track_count} tracks, but only"
                                                        f"{len(entry_config_list)} items in input.")

        # append entry config
        self.lock.writer_enter()
        entry_idx_list = []
        for track_idx in range(self.track_count):
            track = self.track_list[track_idx]
            _, entry_idx = track.append_entry_config(entry_config=entry_config_list[track_idx])
            entry_idx_list.append(entry_idx)
        self.lock.writer_leave()

        # check data integrity
        if not len(set(entry_idx_list)) == 1:
            raise SMOS_exceptions.SMOSTrackUnaligned("Track unaligned.")

        # return
        return SMOS_SUCCESS, entry_idx_list[0]

    # read
    def read_entry_config(self, idx):
        """
        Read entry config at given index and add read reference to that entry.

        :exception SMOS_exceptions.SMOSTrackUnaligned: if some tracks return
                   index out of range while others do not

        :param idx: index of entry to be read
        :return: [SMOS_SUCCESS, entry_config_list] if successful,
                 [SMOS_FAIL, None] if target entry does not exist
        """
        # read entry config
        status_list = []
        entry_config_list = []
        self.lock.reader_enter()
        for track in self.track_list:
            status, entry_config = track.read_entry_config(idx=idx)
            status_list.append(status)
            entry_config_list.append(entry_config)
        self.lock.reader_leave()

        # check data integrity
        if not len(set(status_list)) == 1:
            raise SMOS_exceptions.SMOSTrackUnaligned("Track unaligned.")

        # return
        if status_list[0] == SMOS_SUCCESS:
            return SMOS_SUCCESS, entry_config_list
        else:
            return SMOS_FAIL, None

    def release_read_reference(self, idx):
        """
        Release read reference on given entry

        :exception SMOS_exceptions.SMOSTrackUnaligned: if some tracks return
                   index out of range while others do not

        :param idx: index of entry to be released
        :return: SMOS_SUCCESS if successful,
                 SMOS_FAIL if target entry does not exist
        """
        # release read reference
        status_list = []
        self.lock.reader_enter()
        for track in self.track_list:
            status = track.release_read_reference(idx=idx)
            status_list.append(status)
        self.lock.reader_leave()

        # check integrity and return
        if not len(set(status_list)) == 1:
            raise SMOS_exceptions.SMOSTrackUnaligned("Track unaligned.")
        return status_list[0]

    # delete
    def delete_entry_config(self, idx, force_delete=False):
        """
        Delete entry at idx from current SharedMemoryObject. Note that this is lazy
        delete, the actual data in shared memory is not erased.

        :exception SMOS_exceptions.SMOSTrackUnaligned: if different tracks have different
                   return value to delete operation

        :param idx: index of entry to be deleted
        :param force_delete: whether to delete the entry when there are still pending readers
        :return: SMOS_SUCCESS if successful,
                 SMOS_FAIL if target entry does not exist,
                 SMOS_PERMISSION_DENIED if permission denied
        """
        # delete entry config
        status_list = []
        self.lock.writer_enter()
        for track in self.track_list:
            status = track.delete_entry_config(idx=idx, force_delete=force_delete)
            status_list.append(status)
        self.lock.writer_leave()

        # check integrity and return
        if not len(set(status_list)) == 1:
            raise SMOS_exceptions.SMOSTrackUnaligned("Track unaligned.")
        return status_list[0]

    # pop and free
    def pop_entry_config(self, force_pop=False):
        """
        Pop an entry from SharedMemoryObject. Note that blocks to which the entry is mapped
        will not be freed in this function (since data in the entry will be used after pop).
        Call free_block_mapping to free the block when data in the entry is no longer useful.

        :exception SMOS_exceptions.SMOSTrackUnaligned: if different tracks have different
                   return value to delete operation

        :param force_pop: whether to pop the entry when there are still pending readers
        :return: [SMOS_SUCCESS, entry_config_list] if successful,
                 [SMOS_FAIL, None] if data track empty,
                 [SMOS_PERMISSION_DENIED, None] if permission denied
        """
        # pop entry config
        status_list = []
        entry_config_list = []
        self.lock.writer_enter()
        for track in self.track_list:
            status, entry_config = track.pop_entry_config(force_pop=force_pop)
            status_list.append(status)
            entry_config_list.append(entry_config)
        self.lock.writer_leave()

        # check integrity
        if not len(set(status_list)) == 1:
            raise SMOS_exceptions.SMOSTrackUnaligned("Track unaligned.")

        # return
        if status_list[0] == SMOS_SUCCESS:
            return SMOS_SUCCESS, entry_config_list
        else:
            return status_list[0], None

    def free_block_mapping(self, entry_config_list: [utils.EntryConfig]):
        """
        Free blocks associated with a previously popped entry.

        :param entry_config_list: configurations of entry to be freed
        :return: always SMOS_SUCCESS
        """
        # check input shape
        if not len(entry_config_list) == self.track_count:
            raise SMOS_exceptions.SMOSDimensionMismatch(f"There are {self.track_count} tracks, but only"
                                                        f"{len(entry_config_list)} items in input.")

        # free block mapping and return
        for track_idx in range(self.track_count):
            self.track_list[track_idx].free_block_mapping(entry_config=entry_config_list[track_idx])
        return SMOS_SUCCESS

    # stop SharedMemoryObject
    def stop(self):
        """
        Stop SharedMemoryObject for safe exit.

        :return: always SMOS_SUCCESS
        """
        # stop tracks in this object
        for track in self.track_list:
            track.stop()

        # release lock
        del self.lock

    # utility functions for SharedMemoryObject
    def get_shm_name_list(self):
        """
        Get name of shared memories used in this object.

        :return: list of shared memory names
        """
        return self.shm_name_list

    def get_entry_offset(self, entry_config_list: [utils.EntryConfig]):
        """
        Get offset of each track for given entry in shared memory space.

        :exception SMOS_exceptions.SMOSDimensionMismatch: if length of input is
                   different of number of tracks

        :param entry_config_list: configurations of entry to be queried
        :return: always [SMOS_SUCCESS, offset_list]
        """
        # check input shape
        if not len(entry_config_list) == self.track_count:
            raise SMOS_exceptions.SMOSDimensionMismatch(f"There are {self.track_count} tracks, but only"
                                                        f"{len(entry_config_list)} items in input.")

        # get entry offset
        offset_list = []
        for track_idx in range(self.track_count):
            _, offset = self.track_list[track_idx].get_entry_offset(entry_config=entry_config_list[track_idx])
            offset_list.append(offset)

        # return
        return SMOS_SUCCESS, offset_list

    def get_entry_idx_list(self):
        """
        Get index of all entries in current SharedMemoryObject.

        :return: entry_idx_list
        """
        # get entry_idx_list
        self.lock.reader_enter()
        entry_idx_list = self.track_list[0].entry_config_list.keys()
        self.lock.reader_leave()

        # return
        return entry_idx_list

    def get_entry_count(self):
        """
        Get number of entries in current SharedMemoryObject. Note that this
        function also checks data integrity (i.e. whether all tracks are
        properly aligned).

        :exception SMOS_exceptions.SMOSTrackUnaligned: if tracks have different
                   number of entries in them

        :return: entry count
        """
        # get entry count
        entry_count_list = []
        self.lock.reader_enter()
        for track in self.track_list:
            entry_count = track.get_entry_count()
            entry_count_list.append(entry_count)
        self.lock.reader_leave()

        # check data integrity and return
        if not len(set(entry_count_list)) == 1:
            raise SMOS_exceptions.SMOSTrackUnaligned("Track unaligned.")
        return entry_count_list[0]
